#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <descrip.h>
#include <lib$routines.h>

#define SOCKET int
#include <mysql.h>

#define MAX_LEN 32000

struct b_param
{
    int n;
    struct dsc$descriptor_d **bstr;
    MYSQL_BIND *bind;
};

static char *cstr(struct dsc$descriptor_d *s)
{
    char *s2 = malloc(s->dsc$w_length + 1);
    memcpy(s2, s->dsc$a_pointer, s->dsc$w_length);
    s2[s->dsc$w_length] = 0;
    return s2;
}

static void bstr(char *buf, int buflen, struct dsc$descriptor_d *d)
{
    lib$sget1_dd(&buflen, d);
    d->dsc$b_dtype = DSC$K_DTYPE_T;
    memcpy(d->dsc$a_pointer, buf, buflen);
}

static MYSQL *realcon(int *con)
{
    return (MYSQL *)(*con);
}

static MYSQL_STMT *realstmt(int *stmt)
{
    return (MYSQL_STMT *)(*stmt);
}

static struct b_param *realparam(int *param)
{
    return (struct b_param *)(*param);
}

void BMYSQL_ERROR(int *con, struct dsc$descriptor_d *msg)
{
   char *temp = (char *)mysql_error(realcon(con));
   bstr(temp, strlen(temp), msg);
}

void BMYSQL_STMT_ERROR(int *stmt, struct dsc$descriptor_d *msg)
{
   char *temp = (char *)mysql_stmt_error(realstmt(stmt));
   bstr(temp, strlen(temp), msg);
}

int BMYSQL_OPEN(struct dsc$descriptor_d *host,
           int *port,
           struct dsc$descriptor_d *usr,
           struct dsc$descriptor_d *pwd,
           struct dsc$descriptor_d *db)
{
    char *host2 = cstr(host);
    char *usr2 = cstr(usr);
    char *pwd2 = cstr(pwd);
    char *db2 = cstr(db);
    MYSQL *con = mysql_init(NULL);
    con = mysql_real_connect(con, host2, usr2, pwd2, db2, *port, NULL, 0);
    free(host2);
    free(usr2);
    free(pwd2);
    free(db2);
    return (int)con;
}

int BMYSQL_STMT_PREPARE(int *con, struct dsc$descriptor_d *sql)
{
    char *sql2 = cstr(sql);
    MYSQL_STMT *stmt = mysql_stmt_init(realcon(con));
    int stat = mysql_stmt_prepare(stmt, sql2, strlen(sql2));
    if(stat != 0) return -1;
    free(sql2);
    return (int)stmt;
}

int BMYSQL_OUTPARAM_INIT(int *nparam)
{
    struct b_param *p = malloc(sizeof(struct b_param));
    p->n = *nparam;
    p->bstr = malloc(*nparam * sizeof(struct dsc$descriptor_d));
    p->bind = malloc(*nparam * sizeof(MYSQL_BIND));
    return (int)p;
}

int BMYSQL_INPARAM_INIT(int *nparam)
{
    struct b_param *p = malloc(sizeof(struct b_param));
    p->n = *nparam;
    p->bstr = NULL;
    p->bind = malloc(*nparam * sizeof(MYSQL_BIND));
    return (int)p;
}

void BMYSQL_OUTPARAM_LONG(int *param, int *ix, long int *val)
{
    struct b_param *p = realparam(param);
    p->bstr[*ix] = NULL; 
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    p->bind[*ix].buffer_type = MYSQL_TYPE_LONG;
    p->bind[*ix].buffer = (char *)val;
    p->bind[*ix].buffer_length = sizeof(long int);
}

void BMYSQL_INPARAM_LONG(int *param, int *ix, long int *val)
{
    struct b_param *p = realparam(param);
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    p->bind[*ix].buffer_type = MYSQL_TYPE_LONG;
    p->bind[*ix].buffer = (char *)val;
    p->bind[*ix].buffer_length = sizeof(long int);
}
void BMYSQL_OUTPARAM_QUAD(int *param, int *ix, long long int *val)
{
    struct b_param *p = realparam(param);
    p->bstr[*ix] = NULL;
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    p->bind[*ix].buffer_type = MYSQL_TYPE_LONGLONG;
    p->bind[*ix].buffer = (char *)val;
    p->bind[*ix].buffer_length = sizeof(long long int);
}

void BMYSQL_INPARAM_QUAD(int *param, int *ix, long long int *val)
{
    struct b_param *p = realparam(param);
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    p->bind[*ix].buffer_type = MYSQL_TYPE_LONGLONG;
    p->bind[*ix].buffer = (char *)val;
    p->bind[*ix].buffer_length = sizeof(long long int);
}

void BMYSQL_OUTPARAM_SFLOAT(int *param, int *ix, float *val)
{
    struct b_param *p = realparam(param);
    p->bstr[*ix] = NULL;
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    p->bind[*ix].buffer_type = MYSQL_TYPE_FLOAT;
    p->bind[*ix].buffer = (char *)val;
    p->bind[*ix].buffer_length = sizeof(float);
}

void BMYSQL_INPARAM_SFLOAT(int *param, int *ix, float *val)
{
    struct b_param *p = realparam(param);
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    p->bind[*ix].buffer_type = MYSQL_TYPE_FLOAT;
    p->bind[*ix].buffer = (char *)val;
    p->bind[*ix].buffer_length = sizeof(float);
}

void BMYSQL_OUTPARAM_TFLOAT(int *param, int *ix, double *val)
{
    struct b_param *p = realparam(param);
    p->bstr[*ix] = NULL;
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    p->bind[*ix].buffer_type = MYSQL_TYPE_DOUBLE;
    p->bind[*ix].buffer = (char *)val;
    p->bind[*ix].buffer_length = sizeof(double);
}

void BMYSQL_INPARAM_TFLOAT(int *param, int *ix, double *val)
{
    struct b_param *p = realparam(param);
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    p->bind[*ix].buffer_type = MYSQL_TYPE_DOUBLE;
    p->bind[*ix].buffer = (char *)val;
    p->bind[*ix].buffer_length = sizeof(double);
}

void BMYSQL_OUTPARAM_STRING(int *param, int *ix, struct dsc$descriptor *val)
{

    struct b_param *p = realparam(param);
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    if(val->dsc$b_class == DSC$K_CLASS_D)
    {
        p->bstr[*ix] = (struct dsc$descriptor_d *)val;
        p->bind[*ix].buffer_type = MYSQL_TYPE_STRING;
        p->bind[*ix].buffer = malloc(MAX_LEN);
        p->bind[*ix].buffer_length = MAX_LEN;
        p->bind[*ix].length = (unsigned long *)malloc(sizeof(unsigned long));
   }
    if(val->dsc$b_class == DSC$K_CLASS_S)
    {
        p->bstr[*ix] = NULL;
        p->bind[*ix].buffer_type = MYSQL_TYPE_STRING;
        p->bind[*ix].buffer = val->dsc$a_pointer;
        p->bind[*ix].buffer_length = val->dsc$w_length;
   }
}

void BMYSQL_INPARAM_STRING(int *param, int *ix, struct dsc$descriptor *val)
{
    struct b_param *p = realparam(param);
    memset(&p->bind[*ix], 0, sizeof(MYSQL_BIND));
    p->bind[*ix].buffer_type = MYSQL_TYPE_STRING;
    p->bind[*ix].buffer = val->dsc$a_pointer;
    p->bind[*ix].buffer_length = val->dsc$w_length;
}

int BMYSQL_STMT_EXECUTE(int *stmt, int *inparam, int *outparam)
{
    int stat;
    struct b_param *pin = realparam(inparam);
    struct b_param *pout = realparam(outparam);
    if(*inparam != 0)
    {
        stat = mysql_stmt_bind_param(realstmt(stmt), pin->bind);
        if(stat != 0) return -1;
    }
    stat = mysql_stmt_execute(realstmt(stmt));
    if(stat != 0) return -1;
    if(*outparam != 0)
    {
        stat = mysql_stmt_bind_result(realstmt(stmt), pout->bind);
        if(stat != 0) return -1;
        stat = mysql_stmt_store_result(realstmt(stmt));
        if(stat != 0) return -1;
    }
    return stat;
}

int BMYSQL_STMT_FETCH(int *stmt, int *param)
{
    struct b_param *p = realparam(param);
    int stat = mysql_stmt_fetch(realstmt(stmt));
    for(int i = 0; i < p->n; i++)
    {
        if(p->bind[i].buffer_type == MYSQL_TYPE_STRING && p->bstr[i] != NULL)
        {
            char *buf = (char *)p->bind[i].buffer;
            int buflen = (int)*(p->bind[i].length);
            bstr(buf, buflen, p->bstr[i]);
        }
    }
    return stat;
}

void BMYSQL_OUTPARAM_FREE(int *param)
{
    struct b_param *p = realparam(param);
    for(int i = 0; i < p->n; i++) 
    {
        if(p->bind[i].buffer_type == MYSQL_TYPE_STRING && p->bstr[i] != NULL)
        {
           free(p->bind[i].buffer);
           free(p->bind[i].length);
        }
    }
    free(p->bind);
    free(p->bstr);
    free(p);
}

void BMYSQL_INPARAM_FREE(int *param)
{
    struct b_param *p = realparam(param);
    free(p->bind);
    free(p);
}

void BMYSQL_STMT_FREE(int *stmt)
{
    mysql_stmt_free_result(realstmt(stmt));
}

void BMYSQL_CLOSE(int *con)
{
    mysql_close(realcon(con));
}

